 
#include <stdint.h> 
#include <stdio.h> 
#include <string.h>

#include "lptport.h"
#include "cn302.h"

//#define DEBUG_CN302

#ifdef DEBUG_CN302
#define dbgprintf(_x_)	printf _x_
#else
#define dbgprintf(_x_)
#endif

static int cn302_clockdelay,cn302_cmddelay,cn302_syncdelay;
static cn302_progress_callback_t cn302_progress_callback=NULL;

/*
	internal functions
*/


static void waitus(int us)
{
// we need some sleep, that is faster than a Millisecond...
// 
// Inputting or outputting any byte from/to port 0x80 should wait 
// for almost exactly 1 microsecond independent of your processor
// type and speed. You can do this multiple times to wait a few 
// microseconds. The port output should have no harmful side effects
// on any standard machine (and some kernel drivers use it).

	while(us--)
	{
		lptport_getstatus_busy();
	}

}

#define OUTPUTIDLEVALUE	1

static uint8_t read_byte(void)
{
	uint8_t buffer = 0;
	uint8_t inbit;
	int pos;
	for(pos = 0; pos <= 7; pos++){
		lptport_setcontrol_strobe(1);	// clock lo
//		waitus(cn302_clockdelay);
		inbit=lptport_getstatus_busy()^1;
		buffer = ((inbit << pos) | buffer);
		lptport_setcontrol_strobe(0);	// clock hi
//		waitus(cn302_clockdelay);
	}
	waitus(cn302_clockdelay);
	lptport_setcontrol_strobe(1);	// clock lo
	return buffer;
}


static void write_byte(uint8_t buffer)
{
	int pos;
	for(pos = 0; pos <= 7; pos++){
		lptport_setcontrol_strobe(1);	// clock lo
//		waitus(cn302_clockdelay);
		lptport_setcontrol_autofeed(((buffer >> pos) & 0x1)^1);
		lptport_setcontrol_strobe(0);	// clock hi
//		waitus(cn302_clockdelay);
	}
	waitus(cn302_clockdelay);
	lptport_setcontrol_strobe(1);	// clock lo
	lptport_setcontrol_autofeed(OUTPUTIDLEVALUE);
}		


static void write_command(uint8_t *buffer)
{
	int x;
	for(x = 0; x < 10; x++)
	{
		write_byte(buffer[x]);
	}
	waitus(cn302_cmddelay);
}

static void out_of_sync(void)
{
	if(cn302_resync() == -1)
	{
		waitus(cn302_syncdelay);
		cn302_resync();
	}
}

/*
	exported functions
*/

int cn302_init(int clockdelay,int cmddelay,int syncdelay)
{
	cn302_clockdelay=clockdelay;
	cn302_cmddelay=cmddelay;
	cn302_syncdelay=syncdelay;
	
	/* enable the interface
	   this is for interfaces with a real busdriver that
	   take +5V VCC from datapins 1-3
	*/
	lptport_setdata(0x07);
	
	if (cn302_sync_queue(64*16) == -1)
	{
		return -1;
	}
	return 1;
}

void cn302_set_progresscallback(cn302_progress_callback_t callback)
{
	 cn302_progress_callback=callback;
}

int cn302_close(void)
{
	// disable the interface
	return lptport_setdata(0);
}

int cn302_sync_queue(int syncbits)
{
	unsigned short buffer = 0;
	uint8_t inbit;
        int pos;
	
	lptport_setcontrol_autofeed(OUTPUTIDLEVALUE);
	
	for(pos = 0; pos < syncbits; pos++)
	{
		lptport_setcontrol_strobe(1);	// clock lo
		waitus(cn302_clockdelay);
		inbit=lptport_getstatus_busy()^1;
		buffer = buffer >> 1;
		buffer = ((inbit << 15) | buffer);
		lptport_setcontrol_strobe(0);	// clock hi
		waitus(cn302_clockdelay);
		if (buffer == 0xEEEE) break;
	}
	lptport_setcontrol_strobe(1);	// clock lo
	
	if (pos == syncbits) 
	{
		return -1;
	}
	return 0;	
}

int cn302_peek_byte(int offset, uint8_t *val)
{
	uint8_t command[] = {0xFF,0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00};
	int tries=0;
	
	*val=0;
retry:	
	command[5] = (char) ((offset >> 16) & 0xFF);
	command[2] = (char) ((offset >> 8) & 0xFF);
	command[3] = (char) (offset & 0xFF);
	write_command(command);
	uint8_t read_one = read_byte();
	uint8_t read_two = read_byte();
	if ((read_one | read_two)){
		tries++;
	//	dbgprintf(("cn302_peek_byte out of sync, try %d : %02x %02x\n",tries,read_one,read_two));
		out_of_sync();
		if (tries<3) 
		{
			goto retry;
		}
		else
		{
			return -1;
		}
	}
	
	*val=read_byte();
			
	return 1;
}

int cn302_poke_byte(int offset, uint8_t val1)
{
	uint8_t command[] = {0xFE,0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1, 0x00};
	uint8_t receive[4];
	int tries=0;
	int y;
retry:
		command[5] = (char) ((offset >> 16) & 0xFF);
		command[2] = (char) ((offset >> 8) & 0xFF);
		command[3] = (char) (offset & 0xFF);
		command[4] = val1;
		command[6] = 0;
		write_command(command);
		
		
		for (y = 0; y < 3; y++){
			receive[y] = read_byte();
		}
		
		if (!(receive[0] | receive[1]) & (receive[2] == val1))
		{
		}
		else
		{
			tries++;
			out_of_sync();
			//dbgprintf(("cn302_poke_byte out of sync, try %d : %02x %02x %02x\n", tries,receive[0],receive[1],receive[2]));
			if (tries<3) 
			{
				goto retry;
			}
			else
			{
				return -1;
			}
			
		}
	return 1;
}


int cn302_read_block(int offset, int size, uint8_t* buffer)
{
	uint8_t command[] = {0xFF,0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00};
	int x = 0, y, rest;
	int tries=0;
	
	if (size >= 0x20){
		for (x = 0; x < (size - (size % 0x20)); x+=0x20){
			tries=0;
retry1:		
			command[5] = (char) ((offset >> 16) & 0xFF);
			command[2] = (char) ((offset >> 8) & 0xFF);
			command[3] = (char) (offset & 0xFF);
			write_command(command);
			uint8_t read_one = read_byte();
			uint8_t read_two = read_byte();
			if ((read_one | read_two)){
				tries++;
				out_of_sync();
				//dbgprintf(("cn302_read_block out of sync, try %d : %02x %02x\n",tries,read_one,read_two));
				if (tries<3) 
				{
					goto retry1;
				}
				else
				{
					return -1;
				}
			}
			for(y = 0; y < 0x20; y++){
				buffer[x+y] = read_byte();
			}
		
			//if(cn302_progress_callback) 
			//	cn302_progress_callback(x,size);
		
			offset += 0x20;
		}
	}
	rest = (size % 0x20);
	if (rest){
		
		tries=0;
retry2:		
		command[5] = (char) ((offset >> 16) & 0xFF);
		command[2] = (char) ((offset >> 8) & 0xFF);
		command[3] = (char) (offset & 0xFF);
		command[8] = rest;
		write_command(command);
		uint8_t read_one = read_byte();
		uint8_t read_two = read_byte();
		if ((read_one | read_two)){
			tries++;
			out_of_sync();
			//dbgprintf(("cn302_read_block out of sync, try %d : %02x %02x\n",tries,read_one,read_two));
			if (tries<3) 
			{
				goto retry2;
			}
			else
			{
				return -1;
			}
		}
		for(y = 0; y < rest; y++){
			buffer[x+y] = read_byte();
		}
		if(cn302_progress_callback) cn302_progress_callback(x,size);
	}
	return 1;		
}

int cn302_write_block(int offset, int size, uint8_t* buffer)
{
	uint8_t command[] = {0xFE,0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x2, 0x00};
	uint8_t receive[4];
	int x;
	int y;
	int tries;

	x=0;y=0;
	
	if(size != 1){
		for (x = 0; x < (size & ~0x1); x+=0x2){
			tries=0;
retry1:			
			command[5] = (char) ((offset >> 16) & 0xFF);
			command[2] = (char) ((offset >> 8) & 0xFF);
			command[3] = (char) (offset & 0xFF);
			command[4] = buffer[x];
			command[6] = buffer[x+1];
			write_command(command);
			for (y = 0; y < 4; y++){
				receive[y] = read_byte();
			}
	
			if ((receive[0] | receive[1]) | (receive[2] != buffer[x]) | (receive[3] != buffer[x+1])){
				tries++;
				out_of_sync();
				//dbgprintf(("cn302_write_block out of sync, try %d : %02x %02x %02x %02x\n",tries, receive[0],receive[1],receive[2],receive[3]));
				if (tries<3) 
				{
					goto retry1;
				}
				else
				{
					return -1;
				}
			}
			
			//if(cn302_progress_callback) cn302_progress_callback(x,size);
			
			offset+=0x2;
		}
	}
	else
	{
		offset+=0x1;
		x--;
	}
	if (size & 0x1){
		offset-=0x1;
		tries=0;
retry2:			
		command[5] = (char) ((offset >> 16) & 0xFF);
		command[2] = (char) ((offset >> 8) & 0xFF);
		command[3] = (char) (offset & 0xFF);
		command[8] = 0x1;
		command[6] = 0;
		command[4] = buffer[x+1];
		write_command(command);
		for (y = 0; y < 3; y++){
			receive[y] = read_byte();
		}
		
		if (!(receive[0] | receive[1]) & (receive[2] == buffer[x+1]))
		{
		}
		else
		{
			tries++;
			out_of_sync();
			//dbgprintf(("cn302_write_block out of sync, try %d : %02x %02x %02x %02x\n", tries,receive[0],receive[1],receive[2],receive[3]));
			if (tries<3) 
			{
				goto retry2;
			}
			else
			{
				return -1;
			}
		}
		if(cn302_progress_callback) cn302_progress_callback(x,size);
		
	}
	return 1;
}


